// TextureRender.hh for fluxbox
// Copyright (c) 2002 Henrik Kinnunen (fluxgen at fluxbox.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

// $Id: TextureRender.hh,v 1.1 2002/11/30 20:07:49 fluxgen Exp $

#ifndef TEXTURRENDER_HH
#define TEXTURRENDER_HH

#include "Texture.hh"

#include <X11/Xlib.h>

class BImageControl;

/**
	Renders texture to pixmap
*/
class TextureRender {
public:
	TextureRender(BImageControl &ic, unsigned int width, unsigned int height, 
		XColor *_colors=0, size_t num_colors=0);
	~TextureRender();
	/// render to pixmap
	Pixmap render(const FbTk::Texture &src_texture);
	/// render solid texture to pixmap
	Pixmap renderSolid(const FbTk::Texture &src_texture);
	/// render gradient texture to pixmap
	Pixmap renderGradient(const FbTk::Texture &src_texture);

private:
	/**
		Render to pixmap
		@return rendered pixmap
	*/
	Pixmap renderPixmap();
	/**
		Render to XImage
		@returns allocated and rendered XImage, user is responsible to deallocate
	*/
	XImage *renderXImage();
	/**
		@name render functions
	*/
	//@{
	void invert();
	void bevel1();
	void bevel2();
	void dgradient();
	void egradient();
	void hgradient();
	void pgradient();
	void rgradient();
	void vgradient();
	void cdgradient();
	void pcgradient();
	//@}
	void makeGradientBuffers();

	BImageControl &control;
	bool interlaced;

	XColor *colors; // color table

	const FbTk::Color *from, *to;
	int red_offset, green_offset, blue_offset, red_bits, green_bits, blue_bits,
		ncolors, cpc, cpccpc;
	unsigned char *red, *green, *blue;
	const unsigned char *red_table, *green_table, *blue_table;
	unsigned int width, height;
	unsigned int *xtable, *ytable;
};

#endif // TEXTURERENDER_HH
